#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <string.h>
#include <pthread.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
/* include local files */
#include "gfx_functions.h"
#include "matricies.cpp"

class GFX_face{
	public:
		MAT_matrix<float> point;
		MAT_matrix<int> texturepoly;
		SDL_Surface *texture;

		GFX_face(){
			texture = (SDL_Surface *)NULL;
		}

		int initialize(int numpoints){
			int returnval;
			if(numpoints > 0){
				returnval = (point.allocate(numpoints, 4) == NULL);
			}else
				returnval = 1;
			return returnval;
		}
};

class GFX_shape{
	public:
		MAT_matrix<float> location;
		GFX_face *face;
		SDL_Surface *texture;
		int numfaces;

		GFX_shape(){
			texture = (SDL_Surface *)NULL;
			face = (GFX_face *)NULL;
		}

		~GFX_shape(){
			if(face != (GFX_face *)NULL) delete face;
		}

		int initialize(int newnumfaces){
			int returnval = 0;
			if(newnumfaces >= 0){
				if(face != NULL){
					delete face;
					numfaces = 0;
				}
				face = new GFX_face[newnumfaces];
				if(face != NULL){
					numfaces = newnumfaces;
				}else{
					numfaces = 0;
					returnval = 2;
				}
			}else
				returnval = 1;
			return returnval;
		}

};

/***** function declarations *****/
int build_box(GFX_shape *myshape);
int build_testshape(GFX_shape *shape);

int main(int argc, char *argv[]){
	int returnval = 0, *depth_plane, n, m, old_mousex = 0, old_mousey = 0, delta;
	GFX_shape myshape;//, testshape[12];
	MAT_matrix<float> transformation, t1;
	MAT_matrix<int> drawpoints, testmat;

	/**** initialize memory de-allocation ****/
	numfreepointers = 0;
	atexit(gfx_free_memory);
	atexit(SDL_Quit);

	/**** initialize graphics and events****/
	returnval |= gfx_init(480, 480, 32, 0 & SDL_FULLSCREEN);
	gfx_initialize_events();

	/*** allocate the Z buffer ***/
	if(!returnval){
		if((depth_plane = (int *)malloc(x_resolution * y_resolution * sizeof(int))) != NULL){
			free_pointer[numfreepointers++] = (void *)depth_plane;
		}
		else returnval = 1;
	}

	/** allocate the transformation matrix **/
	if(!transformation.allocate(4, 4)) returnval = 1;

	/*** build the shape ***/
	returnval |= build_box(&myshape);

	/*** Everything's allocated and initialized successfully ***/
	if(!returnval){
		/*** create out transformation matrix ***/
		transformation = identitymat((float)4.0);

		while(GFX_mousebutton_state != 1){
			gfx_solid_box(screen_image, 0, 0, screen_image->w - 1, screen_image->h - 1, 0, 0, 0);
			memset((void *)depth_plane, 1000, x_resolution * y_resolution * sizeof(int));
			for(n = 0; n < myshape.numfaces; n++){
				myshape.face[n].point = myshape.face[n].point * transformation;


				drawpoints.allocate(myshape.face[n].point.xsize, myshape.face[n].point.ysize);
				for(m = 0; m < myshape.face[n].point.xsize * myshape.face[n].point.ysize; m++){
					drawpoints.value[m] = (int)myshape.face[n].point.value[m] + (y_resolution >> 1);
				}
				gfx_draw_3Dtexture(screen_image, drawpoints.value, myshape.face[n].texturepoly.value, myshape.face[n].point.xsize, myshape.face[n].texture, depth_plane);
			}
			SDL_UpdateRect(screen_image, 0, 0, 0, 0);
			if(GFX_mousebutton_state){
				transformation = identitymat((float)4.0);
			}else if(old_mousex != GFX_mouse_x || old_mousey != GFX_mouse_y){
				old_mousex = GFX_mouse_x;
				old_mousey = GFX_mouse_y;
				transformation = identitymat((float)4.0);
				delta = old_mousex - (x_resolution >> 1);
				transformation *= yrotmat((float)-delta / (float)(x_resolution << 1));
				delta = old_mousey - (y_resolution >> 1);
				transformation *= xrotmat((float)delta / (float)(y_resolution << 1));
			}
		}


	}else{
		fprintf(stderr, "Initialization errors.\n"); fflush(stderr);
	}
	gfx_close_threads();
	return returnval;
}

int build_testshape(GFX_shape *shape){
	int returnval = 0, xsize = 9, ysize = 11;
	int x, y, n, ix, iy;
	float z1, z2, swapcol[4];
	SDL_Surface *texture;

	texture = gfx_load_png("BeachPenguin.png");
	if(texture != NULL){
		free_pointer[numfreepointers++] = (void *)texture;
	}else{
		returnval = 1;
	}

	if(!returnval && !shape->initialize(xsize * ysize)){
		n = 0;
		for(y = 0; y < ysize; y++){
			for(x = 0; x < xsize; x++){
				if(shape->face[n].initialize(4)){
					returnval = 1;
				}else{
					
					z1 = (float)sin((float)y);
					z2 = (float)sin((float)(y + 1));
					swapcol[0] = (float)(x - (xsize >> 1));	
					swapcol[1] = (float)(y - (ysize >> 1));	
					swapcol[2] = z1;
					swapcol[3] = 1.0;
					shape->face[n].point.setcol(0, swapcol);

					swapcol[0] += 1.0;
					shape->face[n].point.setcol(1, swapcol);
					swapcol[1] += 1.0;
					swapcol[2] = z2;
					shape->face[n].point.setcol(2, swapcol);
					swapcol[0] -= 1.0;
					shape->face[n].point.setcol(3, swapcol);
					/*
					for(m = 0; m < 16; m++)
						shape->face[n].point.value[m] = 0;
					*/
					shape->face[n].point *= 50.0;

					if(shape->face[n].texturepoly.allocate(4, 2)){
						ix = texture->w / xsize;
						iy = texture->h / ysize;
						shape->face[n].texturepoly.value[0] = x * ix;
						shape->face[n].texturepoly.value[4] = y * iy;
						shape->face[n].texturepoly.value[1] = (x + 1) * ix;
						shape->face[n].texturepoly.value[5] = y * iy;
						shape->face[n].texturepoly.value[2] = (x + 1) * ix;
						shape->face[n].texturepoly.value[6] = (y + 1) * iy;
						shape->face[n].texturepoly.value[3] = x * ix;
						shape->face[n].texturepoly.value[7] = (y + 1) * iy;

						shape->face[n].texture = texture;
					}else{
						returnval = 1;
					}
				}
				n++;
			}
		}
	}else{
		returnval = 1;
	}

	return returnval;
}

int build_box(GFX_shape *myshape){
	SDL_Surface *texture[6];
	int n, x, y, z, xrange, yrange, returnval = 0;
	MAT_matrix<float> initpoint;

	/**** allocate memory for our shape ****/
	if(!myshape->initialize(6)){
		for(n = 0; n < myshape->numfaces; n++){
			if(myshape->face[n].initialize(4)) returnval = 1;
		}
	}else returnval = 1;

	/*** load textures ***/
	for(n = 0; n < 6; n++){
		switch(n){
			case 0: texture[0] = gfx_load_png("3dpanel.png"); break;
			case 1: texture[1] = gfx_load_png("BeachPenguin.png"); break;
			case 2: texture[2] = gfx_load_png("Sunflower.png"); break;
			case 3: texture[3] = gfx_load_png("mandelbrot.png"); break;
			case 4: texture[4] = gfx_load_png("Martians.png"); break;
			case 5: texture[5] = gfx_load_png("woodpanel.png"); break;
		}
		if(texture != NULL){
			free_pointer[numfreepointers++] = (void *)texture[n];
		}else{
			returnval = 1;
			break;
		}
	}

	/** allocate temporary points for defining our myshape **/
	if(initpoint.allocate(8, 4)){

		/** Define our shape, which will be a cube. **/
		n = 0;
		for(x = -1; x <= 1; x += 2){
			for(y = -1; y <= 1; y += 2){
				for(z = -1; z <= 1; z += 2){
					*initpoint(n, 0) = x;
					*initpoint(n, 1) = y;
					*initpoint(n, 2) = z;
					*initpoint(n, 3) = 1;
					n++;
				}
			}
		}

		myshape->face[0].point.embed(0, 0, initpoint.subset(0, 0, 0, initpoint.ysize - 1));
		myshape->face[0].point.embed(1, 0, initpoint.subset(1, 0, 1, initpoint.ysize - 1));
		myshape->face[0].point.embed(2, 0, initpoint.subset(3, 0, 3, initpoint.ysize - 1));
		myshape->face[0].point.embed(3, 0, initpoint.subset(2, 0, 2, initpoint.ysize - 1));

		myshape->face[1].point.embed(0, 0, initpoint.subset(4, 0, 4, initpoint.ysize - 1));
		myshape->face[1].point.embed(1, 0, initpoint.subset(5, 0, 5, initpoint.ysize - 1));
		myshape->face[1].point.embed(2, 0, initpoint.subset(7, 0, 7, initpoint.ysize - 1));
		myshape->face[1].point.embed(3, 0, initpoint.subset(6, 0, 6, initpoint.ysize - 1));

		myshape->face[2].point.embed(0, 0, initpoint.subset(1, 0, 1, initpoint.ysize - 1));
		myshape->face[2].point.embed(1, 0, initpoint.subset(5, 0, 5, initpoint.ysize - 1));
		myshape->face[2].point.embed(2, 0, initpoint.subset(7, 0, 7, initpoint.ysize - 1));
		myshape->face[2].point.embed(3, 0, initpoint.subset(3, 0, 3, initpoint.ysize - 1));

		myshape->face[3].point.embed(0, 0, initpoint.subset(0, 0, 0, initpoint.ysize - 1));
		myshape->face[3].point.embed(1, 0, initpoint.subset(4, 0, 4, initpoint.ysize - 1));
		myshape->face[3].point.embed(2, 0, initpoint.subset(6, 0, 6, initpoint.ysize - 1));
		myshape->face[3].point.embed(3, 0, initpoint.subset(2, 0, 2, initpoint.ysize - 1));

		myshape->face[4].point.embed(0, 0, initpoint.subset(0, 0, 0, initpoint.ysize - 1));
		myshape->face[4].point.embed(1, 0, initpoint.subset(1, 0, 1, initpoint.ysize - 1));
		myshape->face[4].point.embed(2, 0, initpoint.subset(5, 0, 5, initpoint.ysize - 1));
		myshape->face[4].point.embed(3, 0, initpoint.subset(4, 0, 4, initpoint.ysize - 1));

		myshape->face[5].point.embed(0, 0, initpoint.subset(2, 0, 2, initpoint.ysize - 1));
		myshape->face[5].point.embed(1, 0, initpoint.subset(3, 0, 3, initpoint.ysize - 1));
		myshape->face[5].point.embed(2, 0, initpoint.subset(7, 0, 7, initpoint.ysize - 1));
		myshape->face[5].point.embed(3, 0, initpoint.subset(6, 0, 6, initpoint.ysize - 1));

		/** scale the shape up a bit **/
		for(n = 0; n < 6; n++){
			myshape->face[n].point *= 100.0;
		}
		/*** The faces have all been created, now initialize their textures ***/
		for(n = 0; n < 6; n++){
			/*
			switch(n){
				case 0: x = 0; y = 1; break;
				case 1: x = 2; y = 1; break;
				case 2: x = 1; y = 0; break;
				case 3: x = 1; y = 2; break;
				case 4: x = 1; y = 1; break;
				case 5: x = 1; y = 3; break;
			}
			xrange = (texture->w - 1) >> 2;
			yrange = (texture->h - 1) >> 2;
			*/
			x = y = 0;
			xrange = texture[n]->w - 1;
			yrange = texture[n]->h - 1;
			
			myshape->face[n].texture = texture[n];
			myshape->face[n].texturepoly.allocate(4, 2);

			myshape->face[n].texturepoly.value[0] = x * xrange;
			myshape->face[n].texturepoly.value[4] = y * yrange;
			
			myshape->face[n].texturepoly.value[1] = myshape->face[n].texturepoly.value[0] + xrange;
			myshape->face[n].texturepoly.value[5] = myshape->face[n].texturepoly.value[4];
			
			myshape->face[n].texturepoly.value[2] = myshape->face[n].texturepoly.value[0] + xrange;
			myshape->face[n].texturepoly.value[6] = myshape->face[n].texturepoly.value[4] + yrange;
			
			myshape->face[n].texturepoly.value[3] = myshape->face[n].texturepoly.value[0];
			myshape->face[n].texturepoly.value[7] = myshape->face[n].texturepoly.value[4] + yrange;
		}
	}
	return returnval;
}
